import logging
import abc

import numpy as np
import torch
import typing

from arm_pytorch_utilities.tensor_utils import ensure_tensor
from torch.distributions import MultivariateNormal

from pytorch_mppi import MPPI
# optimizers
import cma

logger = logging.getLogger(__file__)


class EvaluationResult(typing.NamedTuple):
    # (N) cost for each trajectory evaluated
    costs: torch.Tensor
    # (N x H x nx) where H is the horizon and nx is the state dimension
    rollouts: torch.Tensor
    # parameter values populated by the tuner after evaluation returns
    params: dict = None
    # iteration number populated by the tuner after evaluation returns
    iteration: int = None


class Optimizer:
    def __init__(self):
        self.tuner: typing.Optional[AutotuneMPPI] = None
        self.optim = None

    @abc.abstractmethod
    def setup_optimization(self) -> None:
        """Create backend optim object with optimization parameters and MPPI parameters from the tuner"""

    @abc.abstractmethod
    def optimize_step(self) -> EvaluationResult:
        """Optimize a single step, returning the evaluation result from the latest parameters"""

    def optimize_all(self, iterations) -> EvaluationResult:
        """Optimize multiple steps, returning the best evaluation results.
        Some optimizers may only have this implemented."""
        res = None
        for i in range(iterations):
            res = self.optimize_step()
        return res


class CMAESOpt(Optimizer):
    """Optimize using CMA-ES, an evolutionary algorithm that maintains a Gaussian population,
    starting around the initial parameters with a variance (potentially different for each hyperparameter)."""

    def __init__(self, population=10, sigma=0.1):
        self.population = population
        self.sigma = sigma
        super().__init__()

    def setup_optimization(self):
        x0 = self.tuner.flatten_params()

        options = {"popsize": self.population, "seed": np.random.randint(0, 10000), "tolfun": 1e-5, "tolfunhist": 1e-6}
        self.optim = cma.CMAEvolutionStrategy(x0=x0, sigma0=self.sigma, inopts=options)

    def optimize_step(self):
        params = self.optim.ask()
        # convert params for use

        cost_per_param = []
        all_rollouts = []
        for param in params:
            self.tuner.unflatten_params(param)
            res = self.tuner.evaluate_fn()
            cost_per_param.append(res.costs.mean().cpu().numpy())
            all_rollouts.append(res.rollouts)

        cost_per_param = np.array(cost_per_param)
        self.optim.tell(params, cost_per_param)

        best_param = self.optim.best.x
        self.tuner.unflatten_params(best_param)
        res = self.tuner.evaluate_fn()
        return res


class AutotuneMPPI:
    """Tune selected MPPI hyperparameters using state-of-the-art optimizers on an evaluation function.
    Subclass to define other parameters to optimize over such as terminal cost scaling. An example
    evaluate_fn:


    """
    TUNABLE_PARAMS = ['sigma', 'mu', 'lambda', 'horizon']
    eps = 0.0001

    def __init__(self, mppi: MPPI, params_to_tune: typing.Sequence[str],
                 evaluate_fn: typing.Callable[[], EvaluationResult], optimizer=CMAESOpt()):
        self.mppi = mppi
        self.evaluate_fn = evaluate_fn
        self.d = mppi.d
        self.dtype = mppi.dtype

        self.params = None
        self.optim = optimizer
        self.optim.tuner = self
        self.results = []

        self.define_parameters(params_to_tune)
        self.optim.setup_optimization()

    def optimize_step(self) -> EvaluationResult:
        res = self.optim.optimize_step()
        res = self.log_current_result(res)
        return res

    def optimize_all(self, iterations) -> EvaluationResult:
        res = self.optim.optimize_all(iterations)
        res = self.log_current_result(res)
        return res

    def get_best_result(self) -> EvaluationResult:
        return min(self.results, key=lambda res: res.costs.mean().item())

    def log_current_result(self, res: EvaluationResult):
        with torch.no_grad():
            iteration = len(self.results)
            res = res._replace(iteration=iteration,
                               params={k: v.detach().clone() if torch.is_tensor(v) else v for k, v in
                                       self.params.items()})
            logger.info(f"i:{iteration} cost: {res.costs.mean().item()} params:{self.params}")
            self.results.append(res)
        return res

    def define_parameters(self, params_to_tune: typing.Sequence[str]):
        pm = {}
        # take on the assigned values to the MPPI
        if 'sigma' in params_to_tune:
            # we're going to require that sigma be diagonal of positive values to enforce positive definiteness
            pm['sigma'] = torch.cat([self.mppi.noise_sigma[i][i].view(1) for i in range(self.mppi.nu)])
        if 'mu' in params_to_tune:
            pm['mu'] = self.mppi.noise_mu.clone()
        if 'lambda' in params_to_tune:
            pm['lambda'] = self.mppi.lambda_
        if 'horizon' in params_to_tune:
            pm['horizon'] = self.mppi.T
        self.params = pm

    def flatten_params(self):
        x = []
        if 'sigma' in self.params:
            x.append(self.params['sigma'].detach().cpu().numpy())
        if 'mu' in self.params:
            x.append(self.params['mu'].detach().cpu().numpy())
        if 'lambda' in self.params:
            x.append([self.params['lambda']])
        if 'horizon' in self.params:
            x.append([self.params['horizon']])
        x = np.concatenate(x)
        return x

    def unflatten_params(self, x, apply=True):
        # have to be in the same order as the flattening
        params = {}
        nu = self.mppi.nu
        i = 0
        if 'sigma' in self.params:
            sigma = ensure_tensor(self.d, self.dtype, x[i:i + nu])
            sigma[sigma < self.eps] = self.eps
            params['sigma'] = sigma
            i += nu
        if 'mu' in self.params:
            mu = ensure_tensor(self.d, self.dtype, x[i:i + nu])
            params['mu'] = mu
            i += nu
        if 'lambda' in self.params:
            v = max(x[i], self.eps)
            params['lambda'] = v
            i += 1
        if 'horizon' in self.params:
            v = max(round(x[i]), 1)
            params['horizon'] = v
            i += 1
        if apply:
            self.apply_parameters(params)
        return params

    def apply_parameters(self, params):
        if 'sigma' in params:
            # to remain positive definite
            self.mppi.noise_dist = MultivariateNormal(self.mppi.noise_mu, covariance_matrix=torch.diag(params['sigma']))
            self.mppi.noise_sigma_inv = torch.inverse(self.mppi.noise_sigma.detach())
        if 'mu' in params:
            # to remain positive definite
            self.mppi.noise_dist = MultivariateNormal(params['mu'], covariance_matrix=self.mppi.noise_sigma)
            self.mppi.noise_sigma_inv = torch.inverse(self.mppi.noise_sigma.detach())
        if 'lambda' in params:
            self.mppi.lambda_ = params['lambda']
        if 'horizon' in params:
            self.mppi.change_horizon(params['horizon'])
        self.params = params

    def config_to_params(self, config):
        """Configs are param dictionaries where each must be a scalar"""
        nu = self.mppi.nu
        p = self.params

        dtype = self.dtype
        device = self.d

        params = {}
        for name in ['sigma', 'mu']:
            if name in p:
                params[name] = torch.tensor([config[f'{name}{i}'] for i in range(nu)], dtype=dtype, device=device)
        for name in ['lambda', 'horizon']:
            if name in p:
                params[name] = config[name]

        return params



import functools
import logging
import time

import numpy as np
import torch
from torch.distributions.multivariate_normal import MultivariateNormal

logger = logging.getLogger(__name__)


def _ensure_non_zero(cost, beta, factor):
    return torch.exp(-factor * (cost - beta))


def is_tensor_like(x):
    return torch.is_tensor(x) or type(x) is np.ndarray


def squeeze_n(v, n_squeeze):
    for _ in range(n_squeeze):
        v = v.squeeze(0)
    return v


# from arm_pytorch_utilities, standalone since that package is not on pypi yet
def handle_batch_input(n):
    def _handle_batch_input(func):
        """For func that expect 2D input, handle input that have more than 2 dimensions by flattening them temporarily"""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # assume inputs that are tensor-like have compatible shapes and is represented by the first argument
            batch_dims = []
            for arg in args:
                if is_tensor_like(arg):
                    if len(arg.shape) > n:
                        # last dimension is type dependent; all previous ones are batches
                        batch_dims = arg.shape[:-(n - 1)]
                        break
                    elif len(arg.shape) < n:
                        n_batch_dims_to_add = n - len(arg.shape)
                        batch_ones_to_add = [1] * n_batch_dims_to_add
                        args = [v.view(*batch_ones_to_add, *v.shape) if is_tensor_like(v) else v for v in args]
                        ret = func(*args, **kwargs)
                        if isinstance(ret, tuple):
                            ret = [squeeze_n(v, n_batch_dims_to_add) if is_tensor_like(v) else v for v in ret]
                            return ret
                        else:
                            if is_tensor_like(ret):
                                return squeeze_n(ret, n_batch_dims_to_add)
                            else:
                                return ret
            # no batches; just return normally
            if not batch_dims:
                return func(*args, **kwargs)

            # reduce all batch dimensions down to the first one
            args = [v.view(-1, *v.shape[-(n - 1):]) if (is_tensor_like(v) and len(v.shape) > 2) else v for v in args]
            ret = func(*args, **kwargs)
            # restore original batch dimensions; keep variable dimension (nx)
            if type(ret) is tuple:
                ret = [v if (not is_tensor_like(v) or len(v.shape) == 0) else (
                    v.view(*batch_dims, *v.shape[-(n - 1):]) if len(v.shape) == n else v.view(*batch_dims)) for v in
                       ret]
            else:
                if is_tensor_like(ret):
                    if len(ret.shape) == n:
                        ret = ret.view(*batch_dims, *ret.shape[-(n - 1):])
                    else:
                        ret = ret.view(*batch_dims)
            return ret

        return wrapper

    return _handle_batch_input


class MPPI():
    """
    Model Predictive Path Integral control
    This implementation batch samples the trajectories and so scales well with the number of samples K.

    Implemented according to algorithm 2 in Williams et al., 2017
    'Information Theoretic MPC for Model-Based Reinforcement Learning',
    based off of https://github.com/ferreirafabio/mppi_pendulum
    """

    def __init__(self, dynamics, running_cost, nx, noise_sigma, num_samples=100, horizon=15, device="cpu",
                 terminal_state_cost=None,
                 lambda_=1.,
                 noise_mu=None,
                 u_min=None,
                 u_max=None,
                 u_init=None,
                 U_init=None,
                 u_scale=1,
                 u_per_command=1,
                 step_dependent_dynamics=False,
                 rollout_samples=1,
                 rollout_var_cost=0,
                 rollout_var_discount=0.95,
                 sample_null_action=False,
                 noise_abs_cost=False):
        """
        :param dynamics: function(state, action) -> next_state (K x nx) taking in batch state (K x nx) and action (K x nu)
        :param running_cost: function(state, action) -> cost (K) taking in batch state and action (same as dynamics)
        :param nx: state dimension
        :param noise_sigma: (nu x nu) control noise covariance (assume v_t ~ N(u_t, noise_sigma))
        :param num_samples: K, number of trajectories to sample
        :param horizon: T, length of each trajectory
        :param device: pytorch device
        :param terminal_state_cost: function(state) -> cost (K x 1) taking in batch state
        :param lambda_: temperature, positive scalar where larger values will allow more exploration
        :param noise_mu: (nu) control noise mean (used to bias control samples); defaults to zero mean
        :param u_min: (nu) minimum values for each dimension of control to pass into dynamics
        :param u_max: (nu) maximum values for each dimension of control to pass into dynamics
        :param u_init: (nu) what to initialize new end of trajectory control to be; defeaults to zero
        :param U_init: (T x nu) initial control sequence; defaults to noise
        :param step_dependent_dynamics: whether the passed in dynamics needs horizon step passed in (as 3rd arg)
        :param rollout_samples: M, number of state trajectories to rollout for each control trajectory
            (should be 1 for deterministic dynamics and more for models that output a distribution)
        :param rollout_var_cost: Cost attached to the variance of costs across trajectory rollouts
        :param rollout_var_discount: Discount of variance cost over control horizon
        :param sample_null_action: Whether to explicitly sample a null action (bad for starting in a local minima)
        :param noise_abs_cost: Whether to use the absolute value of the action noise to avoid bias when all states have the same cost
        """
        self.d = device
        self.dtype = noise_sigma.dtype
        self.K = num_samples  # N_SAMPLES
        self.T = horizon  # TIMESTEPS

        # dimensions of state and control
        self.nx = nx
        self.nu = 1 if len(noise_sigma.shape) == 0 else noise_sigma.shape[0]
        self.lambda_ = lambda_

        if noise_mu is None:
            noise_mu = torch.zeros(self.nu, dtype=self.dtype)

        if u_init is None:
            u_init = torch.zeros_like(noise_mu)

        # handle 1D edge case
        if self.nu == 1:
            noise_mu = noise_mu.view(-1)
            noise_sigma = noise_sigma.view(-1, 1)

        # bounds
        self.u_min = u_min
        self.u_max = u_max
        self.u_scale = u_scale
        self.u_per_command = u_per_command
        # make sure if any of them is specified, both are specified
        if self.u_max is not None and self.u_min is None:
            if not torch.is_tensor(self.u_max):
                self.u_max = torch.tensor(self.u_max)
            self.u_min = -self.u_max
        if self.u_min is not None and self.u_max is None:
            if not torch.is_tensor(self.u_min):
                self.u_min = torch.tensor(self.u_min)
            self.u_max = -self.u_min
        if self.u_min is not None:
            self.u_min = self.u_min.to(device=self.d)
            self.u_max = self.u_max.to(device=self.d)

        self.noise_mu = noise_mu.to(self.d)
        self.noise_sigma = noise_sigma.to(self.d)
        self.noise_sigma_inv = torch.inverse(self.noise_sigma)
        self.noise_dist = MultivariateNormal(self.noise_mu, covariance_matrix=self.noise_sigma)
        # T x nu control sequence
        self.U = U_init
        self.u_init = u_init.to(self.d)

        if self.U is None:
            self.U = self.noise_dist.sample((self.T,))

        self.step_dependency = step_dependent_dynamics
        self.F = dynamics
        self.running_cost = running_cost
        self.terminal_state_cost = terminal_state_cost
        self.sample_null_action = sample_null_action
        self.noise_abs_cost = noise_abs_cost
        self.state = None

        # handling dynamics models that output a distribution (take multiple trajectory samples)
        self.M = rollout_samples
        self.rollout_var_cost = rollout_var_cost
        self.rollout_var_discount = rollout_var_discount

        # sampled results from last command
        self.cost_total = None
        self.cost_total_non_zero = None
        self.omega = None
        self.states = None
        self.actions = None

    @handle_batch_input(n=2)
    def _dynamics(self, state, u, t):
        return self.F(state, u, t) if self.step_dependency else self.F(state, u)

    @handle_batch_input(n=2)
    def _running_cost(self, state, u):
        return self.running_cost(state, u)

    def command(self, state, shift_nominal_trajectory=True):
        """
        :param state: (nx) or (K x nx) current state, or samples of states (for propagating a distribution of states)
        :param shift_nominal_trajectory: Whether to roll the nominal trajectory forward one step. This should be True
        if the command is to be executed. If the nominal trajectory is to be refined then it should be False.
        :returns action: (nu) best action
        """
        if shift_nominal_trajectory:
            # shift command 1 time step
            self.U = torch.roll(self.U, -1, dims=0)
            self.U[-1] = self.u_init

        return self._command(state)

    def _command(self, state):
        if not torch.is_tensor(state):
            state = torch.tensor(state)
        self.state = state.to(dtype=self.dtype, device=self.d)
        cost_total = self._compute_total_cost_batch()
        beta = torch.min(cost_total)
        self.cost_total_non_zero = _ensure_non_zero(cost_total, beta, 1 / self.lambda_)
        eta = torch.sum(self.cost_total_non_zero)
        self.omega = (1. / eta) * self.cost_total_non_zero
        perturbations = []
        for t in range(self.T):
            perturbations.append(torch.sum(self.omega.view(-1, 1) * self.noise[:, t], dim=0))
        perturbations = torch.stack(perturbations)
        self.U = self.U + perturbations
        action = self.U[:self.u_per_command]
        # reduce dimensionality if we only need the first command
        if self.u_per_command == 1:
            action = action[0]
        return action

    def change_horizon(self, horizon):
        if horizon < self.U.shape[0]:
            # truncate trajectory
            self.U = self.U[:horizon]
        elif horizon > self.U.shape[0]:
            # extend with u_init
            self.U = torch.cat((self.U, self.u_init.repeat(horizon - self.U.shape[0], 1)))
        self.T = horizon

    def reset(self):
        """
        Clear controller state after finishing a trial
        """
        self.U = self.noise_dist.sample((self.T,))

    def _compute_rollout_costs(self, perturbed_actions):
        K, T, nu = perturbed_actions.shape
        assert nu == self.nu

        cost_total = torch.zeros(K, device=self.d, dtype=self.dtype)
        cost_samples = cost_total.repeat(self.M, 1)
        cost_var = torch.zeros_like(cost_total)

        # allow propagation of a sample of states (ex. to carry a distribution), or to start with a single state
        if self.state.shape == (K, self.nx):
            state = self.state
        else:
            state = self.state.view(1, -1).repeat(K, 1)

        # rollout action trajectory M times to estimate expected cost
        state = state.repeat(self.M, 1, 1)

        states = []
        actions = []
        for t in range(T):
            u = self.u_scale * perturbed_actions[:, t].repeat(self.M, 1, 1)
            state = self._dynamics(state, u, t)
            c = self._running_cost(state, u)
            cost_samples = cost_samples + c
            if self.M > 1:
                cost_var += c.var(dim=0) * (self.rollout_var_discount ** t)

            # Save total states/actions
            states.append(state)
            actions.append(u)

        # Actions is K x T x nu
        # States is K x T x nx
        actions = torch.stack(actions, dim=-2)
        states = torch.stack(states, dim=-2)

        # action perturbation cost
        if self.terminal_state_cost:
            c = self.terminal_state_cost(states, actions)
            cost_samples = cost_samples + c
        cost_total = cost_total + cost_samples.mean(dim=0)
        cost_total = cost_total + cost_var * self.rollout_var_cost
        return cost_total, states, actions

    def _compute_total_cost_batch(self):
        # parallelize sampling across trajectories
        # resample noise each time we take an action
        noise = self.noise_dist.rsample((self.K, self.T))
        # broadcast own control to noise over samples; now it's K x T x nu
        perturbed_action = self.U + noise
        if self.sample_null_action:
            perturbed_action[self.K - 1] = 0
        # naively bound control
        self.perturbed_action = self._bound_action(perturbed_action)
        # bounded noise after bounding (some got cut off, so we don't penalize that in action cost)
        self.noise = self.perturbed_action - self.U
        if self.noise_abs_cost:
            action_cost = self.lambda_ * torch.abs(self.noise) @ self.noise_sigma_inv
            # NOTE: The original paper does self.lambda_ * torch.abs(self.noise) @ self.noise_sigma_inv, but this biases
            # the actions with low noise if all states have the same cost. With abs(noise) we prefer actions close to the
            # nomial trajectory.
        else:
            action_cost = self.lambda_ * self.noise @ self.noise_sigma_inv  # Like original paper

        rollout_cost, self.states, actions = self._compute_rollout_costs(self.perturbed_action)
        self.actions = actions / self.u_scale

        # action perturbation cost
        perturbation_cost = torch.sum(self.U * action_cost, dim=(1, 2))
        self.cost_total = rollout_cost + perturbation_cost
        return self.cost_total

    def _bound_action(self, action):
        if self.u_max is not None:
            for t in range(self.T):
                u = action[:, self._slice_control(t)]
                cu = torch.max(torch.min(u, self.u_max), self.u_min)
                action[:, self._slice_control(t)] = cu
        return action

    def _slice_control(self, t):
        return slice(t * self.nu, (t + 1) * self.nu)

    def get_rollouts(self, state, num_rollouts=1):
        """
            :param state: either (nx) vector or (num_rollouts x nx) for sampled initial states
            :param num_rollouts: Number of rollouts with same action sequence - for generating samples with stochastic
                                 dynamics
            :returns states: num_rollouts x T x nx vector of trajectories

        """
        state = state.view(-1, self.nx)
        if state.size(0) == 1:
            state = state.repeat(num_rollouts, 1)

        T = self.U.shape[0]
        states = torch.zeros((num_rollouts, T + 1, self.nx), dtype=self.U.dtype, device=self.U.device)
        states[:, 0] = state
        for t in range(T):
            states[:, t + 1] = self._dynamics(states[:, t].view(num_rollouts, -1),
                                              self.u_scale * self.U[t].view(num_rollouts, -1), t)
        return states[:, 1:]


def run_mppi(mppi, env, retrain_dynamics, retrain_after_iter=50, iter=1000, render=True):
    dataset = torch.zeros((retrain_after_iter, mppi.nx + mppi.nu), dtype=mppi.U.dtype, device=mppi.d)
    total_reward = 0
    for i in range(iter):
        state = env.state.copy()
        command_start = time.perf_counter()
        action = mppi.command(state)
        elapsed = time.perf_counter() - command_start
        s, r, _, _ = env.step(action.cpu().numpy())
        total_reward += r
        logger.debug("action taken: %.4f cost received: %.4f time taken: %.5fs", action, -r, elapsed)
        if render:
            env.render()

        di = i % retrain_after_iter
        if di == 0 and i > 0:
            retrain_dynamics(dataset)
            # don't have to clear dataset since it'll be overridden, but useful for debugging
            dataset.zero_()
        dataset[di, :mppi.nx] = torch.tensor(state, dtype=mppi.U.dtype)
        dataset[di, mppi.nx:] = action
    return total_reward, dataset



import numpy as np

# pip install ribs
import ribs

from pytorch_mppi import autotune
from pytorch_mppi.autotune_global import AutotuneMPPIGlobal


class CMAMEOpt(autotune.Optimizer):
    """Quality Diversity optimize using CMA-ME to find a set of good and diverse hyperparameters"""

    def __init__(self, population=10, sigma=1.0, bins=15):
        """

        :param population: number of parameters to sample at once (scales linearly)
        :param sigma: initial variance along all dimensions
        :param bins: int or a Sequence[int] for each hyperparameter for the number of bins in the archive.
        More bins means more granularity along that dimension.
        """
        self.population = population
        self.sigma = sigma
        self.archive = None
        self.qd_score_offset = -3000
        self.num_emitters = 1
        self.bins = bins
        super().__init__()

    def setup_optimization(self):
        if not isinstance(self.tuner, AutotuneMPPIGlobal):
            raise RuntimeError(f"Quality diversity optimizers require global search space information provided "
                               f"by AutotuneMPPIGlobal")

        x = self.tuner.flatten_params()
        ranges = self.tuner.linearized_search_space()
        ranges = list(ranges.values())

        param_dim = len(x)
        bins = self.bins
        if isinstance(bins, (int, float)):
            bins = [bins for _ in range(param_dim)]
        self.archive = ribs.archives.GridArchive(solution_dim=param_dim,
                                                 dims=bins,
                                                 ranges=ranges,
                                                 seed=np.random.randint(0, 10000), qd_score_offset=self.qd_score_offset)
        emitters = [
            ribs.emitters.EvolutionStrategyEmitter(self.archive, x0=x, sigma0=self.sigma, batch_size=self.population,
                                                   seed=np.random.randint(0, 10000)) for i in
            range(self.num_emitters)
        ]
        self.optim = ribs.schedulers.Scheduler(self.archive, emitters)

    def optimize_step(self):
        if not isinstance(self.tuner, AutotuneMPPIGlobal):
            raise RuntimeError(f"Quality diversity optimizers require global search space information provided "
                               f"by AutotuneMPPIGlobal")

        params = self.optim.ask()
        # measure is the whole hyperparameter set - we want to diverse along each dimension

        cost_per_param = []
        all_rollouts = []
        bcs = []
        for param in params:
            full_param = self.tuner.unflatten_params(param)
            res = self.tuner.evaluate_fn()
            cost_per_param.append(res.costs.mean().cpu().numpy())
            all_rollouts.append(res.rollouts)
            behavior = self.tuner.linearize_params(full_param).detach().cpu().numpy()
            bcs.append(behavior)

        cost_per_param = np.array(cost_per_param)
        self.optim.tell(-cost_per_param, bcs)

        best_param = self.archive.best_elite
        # best_param = self.optim.best.x
        self.tuner.unflatten_params(best_param.solution)
        res = self.tuner.evaluate_fn()
        return res

    def get_diverse_top_parameters(self, num_top):
        df = self.archive.as_pandas()
        objectives = df.objective_batch()
        solutions = df.solution_batch()
        # store to allow restoring on next step
        if len(solutions) > num_top:
            order = np.argpartition(-objectives, num_top)
            solutions = solutions[order[:num_top]]

        return [self.tuner.unflatten_params(x, apply=False) for x in solutions]



import torch
from pytorch_mppi.mppi import handle_batch_input


@handle_batch_input(n=2)
def add_2d(a, b):
    assert a.ndim == 2
    assert b.ndim == 2
    return a + b


@handle_batch_input(n=3)
def add_3d(a, b):
    assert a.ndim == 3
    assert b.ndim == 3
    return a + b


def test_batch_wrapper_2d():
    a_2d = torch.tensor([[0.1, 0.2, 0.3]])
    b_2d = torch.tensor([[0.5, -0.2, 0.3]])
    a_3d = torch.tile(a_2d, [1, 1, 1])
    b_3d = torch.tile(b_2d, [1, 1, 1])
    a_4d = torch.tile(a_3d, [2, 1, 1])
    b_4d = torch.tile(b_3d, [2, 1, 1])
    expected_sum_2d = torch.tensor([[0.6, 0.0, 0.6]])
    expected_sum_3d = torch.tensor([[[0.6, 0.0, 0.6]]])
    expected_sum_4d = torch.tensor([[[[0.6, 0.0, 0.6]]], [[[0.6, 0.0, 0.6]]]])
    sum_2d = add_2d(a_2d, b_2d)
    sum_3d = add_2d(a_3d, b_3d)
    sum_4d = add_2d(a_4d, b_4d)
    assert torch.allclose(sum_2d, expected_sum_2d)
    assert torch.allclose(sum_3d, expected_sum_3d)
    assert torch.allclose(sum_4d, expected_sum_4d)


def test_batch_wrapper_3d():
    a_3d = torch.tensor([[[0.1, 0.2, 0.3]]])
    b_3d = torch.tensor([[[0.5, -0.2, 0.3]]])
    a_4d = torch.tile(a_3d, [2, 1, 1])
    b_4d = torch.tile(b_3d, [2, 1, 1])
    expected_sum_3d = torch.tensor([[[0.6, 0.0, 0.6]]])
    expected_sum_4d = torch.tensor([[[[0.6, 0.0, 0.6]]], [[[0.6, 0.0, 0.6]]]])
    sum_3d = add_3d(a_3d, b_3d)
    sum_4d = add_3d(a_4d, b_4d)
    assert torch.allclose(sum_3d, expected_sum_3d)
    assert torch.allclose(sum_4d, expected_sum_4d)



import numpy as np
import torch

# pip install "ray[tune]" bayesian-optimization hyperopt
from ray import tune

from pytorch_mppi import autotune
from ray.tune.search.hyperopt import HyperOptSearch


class AutotuneMPPIGlobal(autotune.AutotuneMPPI):
    def __init__(self, *args,
                 sigma_search_space=tune.loguniform(1e-4, 1e2),
                 mu_search_space=tune.uniform(-1, 1),
                 lambda_search_space=tune.loguniform(1e-5, 1e3),
                 horizon_search_space=tune.randint(1, 50),
                 **kwargs):
        self.sigma_search_space = sigma_search_space
        self.mu_search_space = mu_search_space
        self.lambda_search_space = lambda_search_space
        self.horizon_search_space = horizon_search_space
        super().__init__(*args, **kwargs)

    def search_space(self):
        nu = self.mppi.nu
        p = self.params
        space = {}
        if 'sigma' in p:
            space.update({f"sigma{i}": self.sigma_search_space for i in range(nu)})
        if 'mu' in p:
            space.update({f"mu{i}": self.mu_search_space for i in range(nu)})
        if 'lambda' in p:
            space['lambda'] = self.lambda_search_space
        if 'horizon' in p:
            space['horizon'] = self.horizon_search_space
        return space

    def linearized_search_space(self):
        return {k: self._linearize_search_space(space) for k, space in self.search_space().items()}

    def linearize_params(self, params):
        nu = self.mppi.nu
        p = params
        v = []
        if 'sigma' in p:
            v.extend([self._linearize_space_value(self.sigma_search_space, p['sigma'][i].item()) for i in range(nu)])
        if 'mu' in p:
            v.extend([self._linearize_space_value(self.mu_search_space, p['mu'][i].item()) for i in range(nu)])
        if 'lambda' in p:
            v.append(self._linearize_space_value(self.lambda_search_space, p['lambda']))
        if 'horizon' in p:
            v.append(self._linearize_space_value(self.horizon_search_space, p['horizon']))
        return torch.tensor(v, device=self.d, dtype=self.dtype)

    @staticmethod
    def _linearize_search_space(space):
        # tune doesn't have public API for type checking samplers
        sampler = space.get_sampler()
        if hasattr(sampler, 'base'):
            b = np.log(sampler.base)
            return np.log(space.lower) / b, np.log(space.upper) / b
        return space.lower, space.upper

    @staticmethod
    def _linearize_space_value(space, v):
        # tune doesn't have public API for type checking samplers
        sampler = space.get_sampler()
        # log
        if hasattr(sampler, 'base'):
            b = np.log(sampler.base)
            return np.log(v) / b
        # quantized
        if hasattr(sampler, 'q'):
            return np.round(np.divide(v, sampler.q)) * sampler.q
        return v

    def initial_value(self):
        nu = self.mppi.nu
        p = self.params
        init = {}
        if 'sigma' in p:
            init.update({f"sigma{i}": p['sigma'][i].item() for i in range(nu)})
        if 'mu' in p:
            init.update({f"mu{i}": p['mu'][i].item() for i in range(nu)})
        if 'lambda' in p:
            init['lambda'] = p['lambda']
        if 'horizon' in p:
            init['horizon'] = p['horizon']
        return init


class RayOptimizer(autotune.Optimizer):
    def __init__(self, search_alg=HyperOptSearch,
                 default_iterations=100):
        self.iterations = default_iterations
        self.search_alg = search_alg
        self.all_res = None
        super().__init__()

    def setup_optimization(self):
        if not isinstance(self.tuner, AutotuneMPPIGlobal):
            raise RuntimeError(f"Ray optimizers require global search space information provided by AutotuneMPPIGlobal")
        space = self.tuner.search_space()
        init = self.tuner.initial_value()

        hyperopt_search = self.search_alg(points_to_evaluate=[init], metric="cost", mode="min")
        self.optim = tune.Tuner(
            self.trainable,
            tune_config=tune.TuneConfig(
                num_samples=self.iterations,
                search_alg=hyperopt_search,
                metric="cost",
                mode="min",
            ),
            param_space=space,
        )

    def trainable(self, config):
        self.tuner.apply_parameters(self.tuner.config_to_params(config))
        res = self.tuner.evaluate_fn()
        tune.report(cost=res.costs.mean().item())

    def optimize_step(self):
        raise RuntimeError("Ray optimizers only allow tuning of all iterations at once")

    def optimize_all(self, iterations):
        self.iterations = iterations
        self.setup_optimization()
        self.all_res = self.optim.fit()
        self.tuner.apply_parameters(self.tuner.config_to_params(self.all_res.get_best_result().config))
        res = self.tuner.evaluate_fn()
        return res


